//----------------------------------------------------------------------------
// NAME: ShinyMarble
// VERSION: 0.2
// DESCRIPTION: Renders a 'shiny marble' style Earth with light and water reflections. Adds itself as a layer in Layer Manager (key: L). Right click on layer for settings.
// DEVELOPER: Patrick Murris
// WEBSITE: http://www.alpix.com/3d/worldwin
//----------------------------------------------------------------------------
// Based on the Sky plugin, itself based on Bjorn Reppen 'Atmosphere' plugin
// 0.2 Dec  22, 2005	Fixed layer ordering so that it goes over Images. 
//			Increased sphere density to 128 for beter highlight
//			Turns off below 400km
// 0.1 Dec  14, 2005	First light - MUST TURN OFF IMAGES
//----------------------------------------------------------------------------
// Known issues : Layer ordering not yet appropriate. Goes over places names and other layers
//----------------------------------------------------------------------------

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Windows.Forms;
using WorldWind.Renderable;
using WorldWind.Camera;
using WorldWind;
using System.Drawing;
using System.IO;
using System;

namespace Murris.Plugins
{
	/// <summary>
	/// The plugin (main class)
	/// </summary>
	public class ShinyMarble : WorldWind.PluginEngine.Plugin 
	{
		/// <summary>
		/// Name displayed in layer manager
		/// </summary>
		public static string LayerName = "Shiny Marble";

		/// <summary>
		/// Plugin entry point - All plugins must implement this function
		/// </summary>
		public override void Load() 
		{
			ShinyMarbleLayer layer = new ShinyMarbleLayer(LayerName, PluginDirectory, ParentApplication.WorldWindow);
			//ParentApplication.WorldWindow.CurrentWorld.RenderableObjects.ChildObjects.Insert(0,layer);
			ParentApplication.WorldWindow.CurrentWorld.RenderableObjects.Add(layer);
		}

		/// <summary>
		/// Unloads our plugin
		/// </summary>
		public override void Unload() 
		{
			ParentApplication.WorldWindow.CurrentWorld.RenderableObjects.Remove(LayerName);
		}
	}

	/// <summary>
	/// Sky dome
	/// </summary>
	public class ShinyMarbleLayer : RenderableObject
	{
		static string version = "0.2";
		string settingsFileName = "ShinyMarble.ini";
		string pluginPath;
		public World world;
		public DrawArgs drawArgs;
		Texture texture;
		Mesh mesh;
		Form pDialog;

		// default ShinyMarble bitmap
		public string textureFileName = "world.topo.bathy.200406.png";

		/// <summary>
		/// Constructor
		/// </summary>
		public ShinyMarbleLayer(string LayerName, string pluginPath, WorldWindow worldWindow) : base(LayerName)
		{
			this.pluginPath = pluginPath;
			this.world = worldWindow.CurrentWorld;
			this.drawArgs = worldWindow.DrawArgs;
			this.RenderPriority = RenderPriority.SurfaceImages;
			ReadSettings();
		}
		
		/// <summary>
		/// Read saved settings from ini file
		/// </summary>
		public void ReadSettings()
		{
			string line = "";
			try 
			{
				TextReader tr = File.OpenText(Path.Combine(pluginPath, settingsFileName));
				line = tr.ReadLine();
				tr.Close();
			}
			catch(Exception caught) {}
			if(line != "")
			{
				string[] settingsList = line.Split(';');
				string saveVersion = settingsList[0];	// version when settings where saved
				if(settingsList[1] != null) textureFileName = settingsList[1];
			}
		}

		/// <summary>
		/// Save settings in ini file
		/// </summary>
		public void SaveSettings()
		{
			string line = version + ";" + textureFileName;
			try
			{
				StreamWriter sw = new StreamWriter(Path.Combine(pluginPath, settingsFileName));
				sw.Write(line);
				sw.Close();
			}
			catch(Exception caught) {}
		}

		#region RenderableObject

		/// <summary>
		/// This is where we do our rendering 
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override void Render(DrawArgs drawArgs)
		{
			if(!isInitialized)
				return;

			// Camera & Device shortcuts ;)
			CameraBase camera = drawArgs.WorldCamera;
			Device device = drawArgs.device;

			if(camera.Altitude < 4000e3) return;

			double sphereRadius = camera.WorldRadius + 10e3;
			
			// Create mesh
			if(mesh == null) mesh = TexturedSphere(device, (float)sphereRadius, 128, 128);
			
			// set texture
			device.SetTexture(0,texture);
			device.TextureState[0].ColorOperation = TextureOperation.Modulate;
			
			TextureOperation origAO = device.TextureState[0].AlphaOperation;
			device.TextureState[0].AlphaOperation = TextureOperation.Disable;
			device.VertexFormat = CustomVertex.PositionTextured.Format;
			
			// Set lights
			device.RenderState.Lighting = true;

			Material material = new Material();
			material.Diffuse = Color.White;
			material.Ambient = Color.White;
			material.Specular = Color.White;
			material.SpecularSharpness = 50.0f;  // thin highlight (smaller = larger)
			device.Material = material;
			SetupLights(drawArgs);

			// save world and projection transform
			Matrix origWorld = device.Transform.World;
			Matrix origProjection = device.Transform.Projection;

			// Save fog status
			bool origFog = device.RenderState.FogEnable;
			device.RenderState.FogEnable = false;

			// Set new projection (to avoid being clipped) - probably better ways of doing this?
			float aspectRatio =  (float)device.Viewport.Width / device.Viewport.Height;
			device.Transform.Projection = Matrix.PerspectiveFovRH((float)camera.Fov.Radians, aspectRatio, 1, float.MaxValue );
			// Dont use Z buffer
			device.RenderState.ZBufferEnable = false;

			// draw
			// Shiny layer without alpha (shows water)
			mesh.DrawSubset(0);
			// None shiny layer with alpha (land only)
			device.RenderState.SpecularEnable = false;
			device.TextureState[0].AlphaOperation = origAO;
			mesh.DrawSubset(0);
			

			// Restore device states
			device.Transform.World = origWorld;
			device.Transform.Projection = origProjection;
			device.RenderState.FogEnable = origFog;
			device.RenderState.Lighting = false;
			device.RenderState.ZBufferEnable = true;

		}

		protected void SetupLights(DrawArgs drawArgs)
		{
			Device device = drawArgs.device;
			CameraBase camera = drawArgs.WorldCamera;
			
			// Use a white, directional light 
			device.Lights[0].Diffuse = Color.White;
			device.Lights[0].Specular = Color.FromArgb(0x80, 0x80, 0x80); // soft highlight
			device.Lights[0].Type = LightType.Directional;
			// Compute sun orientation today and now
			int doy = DateTime.UtcNow.DayOfYear;
			int springEq = 80;
			if(doy < springEq) doy += 365;
			double earthInclinaison = 23.5 / 360 * 2 * Math.PI;
			double todayInclinaison = earthInclinaison * Math.Sin((doy - springEq) / 365 * 2 * Math.PI);
			// Setup light direction
			device.Lights[0].Direction = new Vector3((float)Math.Sin(-camera.Longitude.Radians-0.5), (float)Math.Cos(-camera.Longitude.Radians-0.5), -0.5f);
			//device.Lights[0].Direction = new Vector3(3, 3, -1);
			device.Lights[0].Update(); 
			device.Lights[0].Enabled = true;
			device.RenderState.SpecularEnable = true;

			// Add a little ambient light to the scene
			device.RenderState.Ambient = Color.FromArgb(0x40, 0x40, 0x40);
		}

		/// <summary>
		/// RenderableObject abstract member (needed) 
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Initialize(DrawArgs drawArgs)
		{
			try
			{
				texture = TextureLoader.FromFile(drawArgs.device, Path.Combine(pluginPath, textureFileName));
				isInitialized = true;	
			}
			catch
			{
				isOn = false;
				MessageBox.Show("Error loading texture " + Path.Combine(pluginPath, textureFileName) + ".","Layer initialization failed.", MessageBoxButtons.OK, 
					MessageBoxIcon.Error );
			}
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Update(DrawArgs drawArgs)
		{
			if(!isInitialized)
				Initialize(drawArgs);
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Dispose()
		{
			isInitialized = false;
			if(texture!=null)
			{
				texture.Dispose();
				texture = null;
			}
		}

		/// <summary>
		/// Gets called when user left clicks.
		/// RenderableObject abstract member (needed)
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			return false;
		}

 		/// <summary>
 		/// Fills the context menu with menu items specific to the layer.
 		/// </summary>
 		public override void BuildContextMenu( ContextMenu menu )
 		{
  			menu.MenuItems.Add("Properties", new System.EventHandler(OnPropertiesClick));
 		}

 		/// <summary>
 		/// Properties context menu clicked.
 		/// </summary>
 		public void OnPropertiesClick(object sender, EventArgs e)
 		{
			if(pDialog != null && ! pDialog.IsDisposed)
				// Already open
				return;

			// Display the dialog
			pDialog = new propertiesDialog(this);
			pDialog.Show();

 		}

		/// <summary>
		/// Properties Dialog
		/// </summary>
		public class propertiesDialog : System.Windows.Forms.Form
		{
			private System.Windows.Forms.Label lblTexture;
			private System.Windows.Forms.ComboBox cboTexture;
			private System.Windows.Forms.Button btnOK;
			private System.Windows.Forms.Button btnCancel;
			private ShinyMarbleLayer layer;

			public propertiesDialog( ShinyMarbleLayer layer )
			{
				InitializeComponent();
				//this.Icon = WorldWind.PluginEngine.Plugin.Icon;
				this.layer = layer;
				// Init texture list with *.jpg and *.png
				DirectoryInfo di = new DirectoryInfo(layer.pluginPath);
				FileInfo[] imgFiles = di.GetFiles("*.jpg");
				cboTexture.Items.AddRange(imgFiles);
				imgFiles = di.GetFiles("*.png");
				cboTexture.Items.AddRange(imgFiles);
				// select current bitmap
				int i = cboTexture.FindString(layer.textureFileName);
				if(i != -1) cboTexture.SelectedIndex = i;
			}

			#region Windows Form Designer generated code
			/// <summary>
			/// Required method for Designer support - do not modify
			/// the contents of this method with the code editor.
			/// </summary>
			private void InitializeComponent()
			{
				this.btnCancel = new System.Windows.Forms.Button();
				this.btnOK = new System.Windows.Forms.Button();
				this.lblTexture = new System.Windows.Forms.Label();
				this.cboTexture = new System.Windows.Forms.ComboBox();
				this.SuspendLayout();
				// 
				// btnCancel
				// 
				this.btnCancel.DialogResult = System.Windows.Forms.DialogResult.Cancel;
				this.btnCancel.Location = new System.Drawing.Point(311, 59);
				this.btnCancel.Name = "btnCancel";
				this.btnCancel.TabIndex = 0;
				this.btnCancel.Text = "Cancel";
				this.btnCancel.Click += new System.EventHandler(this.btnCancel_Click);
				// 
				// btnOK
				// 
				this.btnOK.Location = new System.Drawing.Point(224, 59);
				this.btnOK.Name = "btnOK";
				this.btnOK.TabIndex = 1;
				this.btnOK.Text = "OK";
				this.btnOK.Click += new System.EventHandler(this.btnOK_Click);
				// 
				// lblTexture
				// 
				this.lblTexture.AutoSize = true;
				this.lblTexture.Location = new System.Drawing.Point(16, 28);
				this.lblTexture.Name = "lblTexture";
				this.lblTexture.Size = new System.Drawing.Size(82, 16);
				this.lblTexture.TabIndex = 2;
				this.lblTexture.Text = "ShinyMarble texture:";
				// 
				// cboTexture
				// 
				this.cboTexture.Location = new System.Drawing.Point(96, 25);
				this.cboTexture.Name = "cboTexture";
				this.cboTexture.Size = new System.Drawing.Size(296, 21);
				this.cboTexture.TabIndex = 3;
				this.cboTexture.Text = "Select texture file";
				this.cboTexture.DropDownStyle = ComboBoxStyle.DropDownList;
				this.cboTexture.MaxDropDownItems = 10;
				// 
				// frmFavorites
				// 
				this.AcceptButton = this.btnOK;
				this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
				this.CancelButton = this.btnCancel;
				this.ClientSize = new System.Drawing.Size(406, 94);
				this.ControlBox = false;
				this.Controls.Add(this.cboTexture);
				this.Controls.Add(this.lblTexture);
				this.Controls.Add(this.btnOK);
				this.Controls.Add(this.btnCancel);
				this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedToolWindow;
				this.MaximizeBox = false;
				this.MinimizeBox = false;
				this.Name = "pDialog";
				this.ShowInTaskbar = false;
				this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
				//this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
				//this.Location = new System.Drawing.Point(layer.drawArgs.CurrentMousePosition.X + 10, layer.drawArgs.CurrentMousePosition.Y - 10);
				this.Text = "ShinyMarble properties";
				this.TopMost = true;
				this.ResumeLayout(false);

			}
			#endregion

			private void btnOK_Click(object sender, System.EventArgs e)
			{
				if(cboTexture.SelectedItem != null) 
				{
  					//System.Windows.Forms.MessageBox.Show("Texture : " + cboTexture.SelectedItem.ToString());
					layer.Dispose();
					layer.textureFileName = cboTexture.SelectedItem.ToString();
					layer.Initialize(layer.drawArgs);
					layer.SaveSettings();
				}
				// Close this form
				this.Close();
			}

			private void btnCancel_Click(object sender, System.EventArgs e)
			{

				// Close this form
				this.Close();
			}





		}

		/// <summary>
		/// Creates a PositionNormalTextured sphere centered on zero
		/// </summary>
		/// <param name="device">The current direct3D drawing device.</param>
		/// <param name="radius">The sphere's radius</param>
		/// <param name="slices">Number of slices (Horizontal resolution).</param>
		/// <param name="stacks">Number of stacks. (Vertical resolution)</param>
		/// <returns></returns>
		/// <remarks>
		/// Number of vertices in the sphere will be (slices+1)*(stacks+1)<br/>
		/// Number of faces	:slices*stacks*2
		/// Number of Indexes	: Number of faces * 3;
		/// </remarks>
		private Mesh TexturedSphere(Microsoft.DirectX.Direct3D.Device device, float radius, int slices, int stacks)
		{
			int numVertices = (slices+1)*(stacks+1);
			int numFaces	= slices*stacks*2;
			int indexCount	= numFaces * 3;

			Mesh mesh = new Mesh(numFaces,numVertices,MeshFlags.Managed,CustomVertex.PositionNormalTextured.Format,device);

			// Get the original sphere's vertex buffer.
			int [] ranks = new int[1];
			ranks[0] = mesh.NumberVertices;
			System.Array arr = mesh.VertexBuffer.Lock(0,typeof(CustomVertex.PositionNormalTextured),LockFlags.None,ranks);

			// Set the vertex buffer
			int vertIndex=0;
			CustomVertex.PositionNormalTextured pnt;
			Vector3 v;
			// south pole vertex
			pnt = new CustomVertex.PositionNormalTextured();
			v = MathEngine.SphericalToCartesian( -90, 180, radius);
			pnt.X = v.X;
			pnt.Y = v.Y;
			pnt.Z = v.Z;
			pnt.Tu = 0.5f;
			pnt.Tv = 1.0f;
			v.Normalize();
			pnt.Normal = v;
			arr.SetValue(pnt,vertIndex++);
			// stacks and slices
			for(int stack = 1; stack < stacks; stack++)
			{
				double latitude = -90 + ((float)stack/stacks*(float)180.0);
				for(int slice = 0; slice <= slices; slice++)
				{
					pnt = new CustomVertex.PositionNormalTextured();
					double longitude = 180 - ((float)slice/slices*(float)360);
					v = MathEngine.SphericalToCartesian( latitude, longitude, radius);
					pnt.X = v.X;
					pnt.Y = v.Y;
					pnt.Z = v.Z;
					pnt.Tu = 1.0f-(float)slice/slices;
					pnt.Tv = 1.0f-(float)stack/stacks;
					v.Normalize();
					pnt.Normal = v;
					arr.SetValue(pnt,vertIndex++);
				}
			}
			// north pole vertex
			pnt = new CustomVertex.PositionNormalTextured();
			v = MathEngine.SphericalToCartesian( 90, 180, radius);
			pnt.X = v.X;
			pnt.Y = v.Y;
			pnt.Z = v.Z;
			pnt.Tu = 0.5f;
			pnt.Tv = 0.0f;
			v.Normalize();
			pnt.Normal = v;
			arr.SetValue(pnt,vertIndex++);

			mesh.VertexBuffer.Unlock();
			ranks[0]=indexCount;
			arr = mesh.LockIndexBuffer(typeof(short),LockFlags.None,ranks);
			int i=0;
			// wrap south pole
			short bottomVertex = 0;
			short topVertex = 1;
			for(int y = 0; y < slices; y++)
			{
				arr.SetValue(bottomVertex,i++);
				arr.SetValue(topVertex,i++);
				arr.SetValue((short)(topVertex+1),i++);
				topVertex++;
			}
			
			// stacks and slices
			for(short x = 1; x < stacks-1; x++)
			{
				bottomVertex = (short)((slices+1) * (x-1) + 1);
				topVertex = (short)(bottomVertex + slices + 1);
				for(int y = 0; y < slices; y++)
				{
					arr.SetValue(bottomVertex,i++);
					arr.SetValue(topVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue(bottomVertex,i++);
					arr.SetValue((short)(topVertex+1),i++);
					arr.SetValue((short)(bottomVertex+1),i++);
					bottomVertex++;
					topVertex++;
				}
			}
			// wrap north pole
			bottomVertex = (short)((slices+1) * (stacks-2) + 1);
			topVertex = (short)(bottomVertex + slices + 1);
			for(int y = 0; y < slices; y++)
			{
				arr.SetValue(bottomVertex,i++);
				arr.SetValue(topVertex,i++);
				arr.SetValue((short)(bottomVertex+1),i++);
				bottomVertex++;
			}


			mesh.IndexBuffer.SetData(arr,0,LockFlags.None);
			//mesh.ComputeNormals();

			return mesh;
		}


		#endregion
	}
}
